"""Spike encoding methods.

Convert continuous-valued signals into spike trains for processing
by spiking neural networks.
"""

from mlxsnn.encoding.rate import rate_encode
from mlxsnn.encoding.latency import latency_encode
from mlxsnn.encoding.delta import delta_encode
from mlxsnn.encoding.medical import EEGEncoder
