"""Medical signal spike encoding methods.

Specialized encoders for converting biomedical signals (EEG, fMRI,
and other biosignals) into spike trains for spiking neural networks.
"""

from mlxsnn.encoding.medical.eeg import EEGEncoder
